import torch
import pandas as pd
import numpy as np
from typing import Sequence, Union, Iterable, Optional
from ._conversions import floats_to_tensor, tensor_to_floats
from ._conversions import named_to_tensor, tensor_to_dataframe
from ._neural_network import NeuralNetwork
from ._loss_functions import MSE
from .layers import Activation, Normalizer


class DenseRegressor(NeuralNetwork):
    """
    DenseRegressor is an implementation of a fully connected NeuralNetwork
    for tabular regression.

    The data type of input 'x' and target 'y' are both a pd.DataFrame, 
    or a dictionary, or a np.ndarray of shape (n_observations, n_classes)

    The prediction output is a np.ndarray of shape (n_observations,)
    if there is a single target, and a pd.DataFrame if there are several
    """

    def __init__(self, inputs: Iterable[str],
                 target: Union[str, Iterable[str]],
                 hidden_layers: Iterable[int],
                 activation: str = "relu",
                 normalize: bool = True,
                 dropout: Optional[float] = None):
        """
        Parameters
        ----------
        inputs : Iterable of str
            the column names of the input variables in a dataframe
        target : str or Iterable of str
            the column name(s) of the variable(s) to predict
        activation : str or Callable or torch.nn.Module
            the activation function
        normalize : bool
            whether or not to add normalization layers
        dropout : float or None
            the dropout after each hidden layer if provided
        """
        super().__init__()
        self.inputs = tuple(inputs)
        self.target = target if isinstance(target, str) else tuple(target)
        self.layers = torch.nn.ModuleList()
        in_features = len(inputs)
        if normalize:
            self.layers.append(Normalizer(in_features, affine=False))
        for out_features in hidden_layers:
            self.layers.append(torch.nn.Linear(in_features, out_features))
            if normalize:
                self.layers.append(Normalizer(out_features))
            self.layers.append(Activation(activation))
            if dropout is not None:
                self.layers.append(torch.nn.Dropout(dropout))
            in_features = out_features
        out_features = 1 if isinstance(target, str) else len(self.target)
        self.output = torch.nn.Linear(in_features, out_features)
        if normalize:
            self.target_norm = Normalizer(out_features, affine=False)
        else:
            self.target_norm = None

    def forward(self, X: torch.Tensor):
        X = X.to(self.device)
        for layer in self.layers:
            X = layer(X)
        return self.output(X)

    def loss(self, x: torch.Tensor, y_target: torch.Tensor,
             weights: Optional[torch.Tensor] = None):
        y_pred = self(x)
        if self.target_norm is not None:
            y_target = self.target_norm(y_target)
        return MSE(y_pred, y_target, weights)

    @property
    def device(self) -> torch.device:
        return self.output.weight.device

    def _x_to_tensor(self, x: Union[pd.DataFrame, dict, Iterable],
                     device: Optional[torch.device] = None):
        return named_to_tensor(x, list(self.inputs), device=device)

    def _y_to_tensor(self, y: Union[pd.DataFrame, dict, Iterable],
                     device: Optional[torch.device] = None) -> torch.Tensor:
        target = [self.target] if isinstance(self.target, str) else list(self.target)
        return named_to_tensor(y, target, device=device)

    def _tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        if self.target_norm is not None:
            tensor = self.target_norm.unscale(tensor)
        if isinstance(self.target, str):
            return tensor_to_floats(tensor).reshape(-1)
        else:
            return tensor_to_dataframe(tensor, self.target)



import requests
import pathlib
from io import IOBase, BytesIO


def download(file_path: str, url: str):
    """
    Download google drive a file from the given url to the disk.
    If the directory does not exists raise an error.
    If the file already exists skip it.

    Parameters
    ----------
    directory : str
        directory in which the file is saved
    file_name : str
        name of the file
    url : str
        url to download it from
    """
    file_path = pathlib.Path(file_path)
    # test if path are valid
    directory = file_path.parent
    if not directory.is_dir():
        raise NotADirectoryError(f"The directory '{directory}' does not exists")
    if file_path.is_file():
        print(f"skipping file '{file_path.name}' as it already exists", flush=True)
        return
    with open(file_path, "wb") as f:
        _download_to_stream(url, f)


def download_bytes(url: str) -> BytesIO:
    """
    Download a google drive file to a BytesIO
    """
    stream = BytesIO()
    _download_to_stream(url, stream)
    stream.seek(0)
    return stream


def _download_to_stream(url: str, stream: IOBase, file_name: str="Download"):
    """
    Download the bytes of the give google drive file into the given IO stream
    """
    # make the request
    session = requests.Session()
    response = session.get(_direct_url(url), stream=True)
    if response.status_code >= 400:
        raise RuntimeError(f"http error: {response.status_code}")
    # get a confirmation token for large files
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            response = session.get(_direct_url(url), params={'confirm': value},
                                   stream=True)
            break
    # save on disk
    CHUNK_SIZE = 32768
    print(f"{file_name}: 0. kB", end="", flush=True)
    for i, chunk in enumerate(response.iter_content(CHUNK_SIZE)):
        stream.write(chunk)
        n_bytes = (i+1)*CHUNK_SIZE / 8.
        for j, unit in enumerate(['Bytes', 'kB', 'MB', 'GB', 'TB']):
            if n_bytes < 1024**(j+1):
                break
        progress = n_bytes / 10024**j
        print(f"\r{file_name}: {progress:.1f} {unit}"+" "*10,
                end="", flush=True)
    print()


def _direct_url(url: str) -> str:
    """
    Converts a googledrive 'share' url to a direct download url

    Parameters
    ----------
    url : str
        the link of of a shared googledrive file

    Returns
    -------
    str :
        the direct download url
    """
    id = url.split("/")[-2]
    return f"https://docs.google.com/uc?export=download&confirm=t&id={id}"



import pathlib
from ._download import download


def titanic(directory: str):
    """downloads 'titanic.csv' in the given directory"""
    download(pathlib.Path(directory) / "titanic.csv",
             "https://drive.google.com/file/d/"
             "1LYjbHW3wyJSMzGMMCmaOFNA_RIKqxRoI/view?usp=sharing")



import torch
import pandas as pd
import numpy as np
from typing import Union, List, Iterable, Tuple, Optional
from .layers import ConvBlock
from ._conversions import tensor_to_classes
from ._conversions import classes_to_tensor, images_to_tensor
from ._conversions import tensor_to_probabilities
from ._neural_network import NeuralNetworkClassifier
from ._loss_functions import cross_entropy


class ImageClassifier(NeuralNetworkClassifier):

    def __init__(self, in_channels: int,
                 classes: Iterable[str],
                 features: Iterable[int],
                 kernel_size: Tuple[int, int] = (3, 3),
                 pooling_size: Optional[Tuple[int, int]] = (2, 2),
                 stride: Tuple[int, int] = (1, 1),
                 activation: str = "relu",
                 n_convs_per_block: int = 1,
                 normalize: bool = True,
                 residuals: bool = True,
                 dropout: Optional[float] = None):
        """
        Parameters
        ----------
        ...
        """
        super().__init__(classes)
        self.layers = torch.nn.ModuleList()
        in_features = in_channels
        for out_features in features:
            self.layers.append(
                ConvBlock(in_features, out_features, kernel_size, stride, activation,
                          normalize, residuals, n_convs_per_block, dropout))
            if pooling_size is not None:
                self.layers.append(torch.nn.MaxPool2d(pooling_size))
            in_features = out_features
        self.output = torch.nn.Linear(out_features, len(self.classes))

    def forward(self, X: torch.Tensor):
        X = X.to(self.device)
        for layer in self.layers:
            X = layer(X)
        N, C, H, W = X.shape
        X = X.reshape(N, C, -1).mean(dim=-1)
        return self.output(X)

    def loss(self, x: torch.Tensor, y_target: torch.Tensor,
             weights: Optional[torch.Tensor] = None,
             class_weights: Optional[torch.Tensor] = None):
        y_pred = self(x)
        return cross_entropy(y_pred, y_target, weights, class_weights)

    @property
    def device(self) -> torch.device:
        return self.output.weight.device

    def _x_to_tensor(self, x: np.ndarray,
                     device: Optional[torch.device] = None):
        return images_to_tensor(x, device=device)

    def _y_to_tensor(self, y: Iterable[str],
                     device: Optional[torch.device] = None) -> torch.Tensor:
        return classes_to_tensor(y, self.classes, device=device)

    def _tensor_to_y(self, tensor: torch.Tensor) -> List[str]:
        return tensor_to_classes(tensor, self.classes)

    def _tensor_to_proba(self, tensor: torch.Tensor) -> pd.DataFrame:
        return tensor_to_probabilities(tensor, self.classes)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Dict, Tuple, Optional
from pygmalion._model_base import ModelBase


class PCA(ModelBase):

    @classmethod
    def from_dump(cls, dump: dict):
        """
        """
        assert dump["type"] == cls.__name__
        projection = {k: (np.array(v[0], dtype=float), v[1]) for k, v in
                      dump["projection"].items()}
        return PCA(dump["offset"], dump["scale"], projection)

    def __init__(self,
                 offset: Dict[str, float] = {}, scale: Dict[str, float] = {},
                 projection: Dict[str, Tuple[np.ndarray, float]] = {}):
        self.offset = pd.Series(offset, dtype=float)
        self.scale = pd.Series(scale, dtype=float)
        self.projection = projection

    def __repr__(self):
        return "PCA("+", ".join(self.projection.keys())+")"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Returns a dataframe of the projection onto the PCA coordinates
        of each observation
        """
        x = df[self.offset.index] - self.offset
        if self.scale is not None:
            x = x/self.scale
        columns = self.projection.keys()
        matrix = np.stack([self.projection[c][0] for c in columns]).T
        pca = x.to_numpy() @ matrix
        return pd.DataFrame(data=pca, columns=columns)

    def fit(self, df: pd.DataFrame, scale: bool = True):
        """
        Compute the principal components of a points cloud

        Parameters
        ----------
        df : pd.DataFrame
            dataframe of numerical values, each column is a variable
        normalize : bool
            If True the data is normalized as (x - mean(x))/std(x)
            If False the data is only translated as (x - mean(x))
        """
        self.offset = df.mean()
        self.scale = df.std() if scale else None
        df = df - self.offset
        if scale:
            df = df/self.scale
        covariance_matrix = np.cov(df.to_numpy().T)
        eigen_values, eigen_vectors = np.linalg.eig(covariance_matrix)
        projections = sorted([(vector, value) for vector, value in
                              zip(eigen_vectors, eigen_values)],
                             key=lambda x: x[1], reverse=True)
        self.projection = {f"PCA{i+1}": (vector, value) for i, (vector, value)
                           in enumerate(projections)}

    def plot_explained_variance(self, ax: Optional[plt.Axes] = None):
        """
        Plot the explained variance as a function of the number of components
        that are kept
        """
        if ax is None:
            f, ax = plt.subplots()
        x = list(range(len(self.projection)+1))
        eigen = [v[1] for v in self.projection.values()]
        total = sum(eigen)
        y = [sum(eigen[:i])/total for i in x]
        ax.plot(x, y)
        ax.set_ylabel("explained variance")
        ax.set_xlabel("number of components")

    @property
    def dump(self):
        offset = dict(self.offset)
        scale = dict(self.scale) if self.scale is not None else self.scale
        projection = {k: [v[0].tolist(), v[1]] for k, v in
                      self.projection.items()}
        return {"type": type(self).__name__,
                "offset": offset,
                "scale": scale,
                "projection": projection}



import pathlib
from ._download import download


def iris(directory: str):
    """downloads 'iris.csv' in the given directory"""
    download(pathlib.Path(directory) / "iris.csv",
             "https://drive.google.com/file/d"
             "/1S1AHfTBtnW1SxsMskRmUcnoDehMbCj0R/view?usp=sharing")



from pygmalion.neural_networks.layers.transformers._attention import KernelizedAttention
import torch
import torch.nn.functional as F
# import pandas as pd
# import matplotlib.pyplot as plt
# from timeit import timeit


def kernel(x):
    return F.elu(x) + 1


def naive_m(q, k, v, RPE, padding_mask=None, scaled=True):
    return KernelizedAttention._kernelized_attention_naive(kernel, q, k, v, True, padding_mask, RPE, scaled=scaled)


def naive_b(q, k, v, RPE, padding_mask=None, scaled=True):
    _, _, Lq, _ = q.shape
    _, _, Lk, _ = k.shape
    return KernelizedAttention._kernelized_attention_naive(kernel, q, k, v, False, padding_mask, RPE, scaled=scaled)


def linear_m(q, k, v, RPE, padding_mask=None, scaled=True):
    return KernelizedAttention._kernelized_attention_linear(kernel, q, k, v, True, padding_mask, RPE, scaled=scaled)


def linear_b(q, k, v, RPE, padding_mask=None, scaled=True):
    _, _, Lq, _ = q.shape
    _, _, Lk, _ = k.shape
    return KernelizedAttention._kernelized_attention_linear(kernel, q, k, v, False, padding_mask, RPE, scaled=scaled)


def test_equality_bidirectional():
    N, H, Lq, Lk, D = 1, 1, 110, 100, 64
    q = torch.rand(N, H, Lq, D)
    k = torch.rand(N, H, Lk, D)
    v = torch.rand(N, H, Lk, D)
    RPE = None
    assert torch.allclose(naive_b(q, k, v, RPE), linear_b(q, k, v, RPE))

def test_equality_masked():
    N, H, Lq, Lk, D = 1, 1, 100, 110, 64
    q = torch.rand(N, H, Lq, D)
    k = torch.rand(N, H, Lk, D)
    v = torch.rand(N, H, Lk, D)
    RPE = None
    assert torch.allclose(naive_m(q, k, v, RPE), linear_m(q, k, v, RPE))

def test_equality_RPE():
    N, H, Lq, Lk, D = 1, 1, 110, 100, 64
    q = torch.rand(N, H, Lq, D)
    k = torch.rand(N, H, Lk, D)
    v = torch.rand(N, H, Lk, D)
    RPE = torch.nn.Embedding(5, D)
    assert torch.allclose(naive_b(q, k, v, RPE), linear_b(q, k, v, RPE))

def test_equality_masked_RPE():
    N, H, Lq, Lk, D = 1, 1, 100, 110, 64
    q = torch.rand(N, H, Lq, D)
    k = torch.rand(N, H, Lk, D)
    v = torch.rand(N, H, Lk, D)
    RPE = torch.nn.Embedding(5, D)
    assert torch.allclose(naive_m(q, k, v, RPE), linear_m(q, k, v, RPE))

def test_equality_padding_masked_RPE():
    N, H, Lq, Lk, D = 1, 1, 100, 110, 64
    q = torch.rand(N, H, Lq, D)
    k = torch.rand(N, H, Lk, D)
    v = torch.rand(N, H, Lk, D)
    padding_mask = (torch.rand(N, Lk) > 0.5)
    RPE = torch.nn.Embedding(5, D)
    assert torch.allclose(naive_m(q, k, v, RPE, padding_mask=padding_mask),
                          linear_m(q, k, v, RPE, padding_mask=padding_mask))

# def benchmark():
#     naive_masked = []
#     naive_bidirectional = []
#     linear_masked = []
#     linear_bidirectional = []

#     n_rep = 10
#     N, H = 1, 1
#     D = 64
#     L = [2**p for p in range(4, 12)]
#     R = 10
#     requires_grad = True
#     device = torch.device("cpu")
#     for l in L:
#         print(l)
#         Lq, Lk = l, l
#         # vectors
#         q = torch.rand((N, H, Lq, D), device=device, requires_grad=requires_grad)
#         v = torch.rand((N, H, Lk, D), device=device, requires_grad=requires_grad)
#         k = torch.rand((N, H, Lk, D), device=device, requires_grad=requires_grad)
#         # attention functions
#         _naive_m = lambda: naive_m(q, k, v)
#         naive_masked.append(timeit(_naive_m, number=n_rep))
#         _naive_b = lambda: naive_b(q, k, v)
#         naive_bidirectional.append(timeit(_naive_b, number=n_rep))
#         _linear_m = lambda: linear_m(q, k, v)
#         linear_masked.append(timeit(_linear_m, number=n_rep))
#         _linear_b = lambda: linear_b(q, k, v)
#         linear_bidirectional.append(timeit(_linear_b, number=n_rep))


#     df = pd.DataFrame(data=zip(L, naive_masked, naive_bidirectional, linear_masked, linear_bidirectional),
#                     columns=["sequences length", "naive masked", "naive bidirectional", "linear masked", "linear bidirectional"])
#     df.to_csv(r"C:\Users\Benoit\Desktop\KA_timing.csv", index=False, encoding="utf-8")


#     plt.style.use("bmh")
#     f, ax = plt.subplots()
#     ax.set_title(f"kernerlized attention runtime for d={D} (best of {n_rep})")
#     ax.set_xscale("log", basex=2)
#     ax.set_yscale("log", basey=2)
#     ax.set_xlabel("Sequences length")
#     ax.set_ylabel("runtime (in seconds)")
#     ax.plot(L, naive_masked, color="C1", linestyle="-", label="naive masked")
#     ax.plot(L, naive_bidirectional, linestyle="--", color="C1", label="naive bidirectional")
#     ax.plot(L, linear_masked, linestyle="-", color="C2", label="linear masked")
#     ax.plot(L, linear_bidirectional, linestyle="--", color="C2", label="linear bidirectional")
#     f.tight_layout()
#     plt.legend()
#     plt.show()


if __name__ == "__main__":
    test_equality_bidirectional()
    test_equality_masked()
    test_equality_RPE()
    test_equality_masked_RPE()
    test_equality_padding_masked_RPE()
    import IPython
    IPython.embed()


import matplotlib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from typing import Iterable, Union, Optional, List


def plot_losses(train_losses: List[float], val_losses: Optional[List[float]] = None,
                grad: Optional[List[float]] = None, best_step: Optional[int] = None,
                ax: Optional[matplotlib.axes.Axes] = None):
    """
    plot the losses
    """
    drawn = []
    if ax is None:
        f, ax = plt.subplots()
    drawn.append(ax.scatter(range(len(train_losses)), train_losses, label="training loss"))
    if val_losses is not None:
        drawn.append(ax.scatter(range(len(val_losses)), val_losses, label="validation loss"))
    ax.set_xlabel("steps")
    ax.set_ylabel("loss")
    ax.set_yscale("log")
    if grad is not None:
        ax2 = ax.twinx()
        drawn.append(ax2.scatter(range(len(grad)), grad, marker=".", color="C2", label="gradient"))
        ax2.set_yscale("log")
        ax.set_zorder(ax2.get_zorder()+1)
        ax.set_frame_on(False)
    if best_step is not None:
        drawn.append(ax.axvline(best_step, 0, 1, color="k", label="best step"))
    ax.legend(drawn, [h.get_label() for h in drawn])


def plot_fitting(x: Iterable[float], y: Iterable[float],
                 ax: Optional[matplotlib.axes.Axes] = None,
                 label: Optional[str] = None,
                 **kwargs):
    """
    Plots the correlation between prediction and target of a regressor

    Parameters
    ----------
    x : iterable of str
        the target to predict
    y : iterable of str
        the prediction of the model
    ax : None or matplotlib.axes.Axes
        The axes to draw on. If None a new window is created.
    label : str
        The legend of the data plotted. Ignored if starts with '_'.
    **kwargs : dict
        dict of keywords passed to 'plt.scatter'
    """
    if ax is None:
        f, ax = plt.subplots(figsize=[5, 5])
    ax.scatter(x, y, label=label, **kwargs)
    points = np.concatenate([c.get_offsets() for c in ax.collections])
    inf, sup = points.min(), points.max()
    delta = sup - inf if sup != inf else 1
    sup += 0.05*delta
    inf -= 0.05*delta
    ax.plot([inf, sup], [inf, sup], color="k", zorder=0)
    ax.set_xlim([inf, sup])
    ax.set_ylim([inf, sup])
    ax.set_aspect("equal", "box")
    legend = ax.legend()
    if len(legend.texts) == 0:
        legend.remove()


def plot_bounding_boxes(bboxes: dict, ax: matplotlib.axes.Axes,
                        class_colors: dict = {}, default_color: str = "r",
                        label_class: bool = True):
    """
    plot the image with given bounding boxes

    Parameters
    ----------
    bounding_boxes : dict
        A dict containing the following keys:
        * x1, y1, x2, y2 : list of int
            The coordinates in pixel of each bboxe corner
        * class : list of str
            The name of the class predicted for eahc bboxe
        * [confidence : list of float]
            The optional confidence of the bounding boxe
    ax : matplotlib.axes.Axes
        The matplotlib axes to draw on
    class_colors : dict
        A dictionary of {class: color} for the color of the boxes
        Can be any color format supported by matplotlib
    default_color : str or list
        the default color for classes that are not present in class_colors
    """
    coords = zip(bboxes["x1"], bboxes["y1"], bboxes["x2"], bboxes["y2"])
    for i, (x1, y1, x2, y2) in enumerate(coords):
        boxe_class = bboxes["class"][i]
        boxe_color = class_colors.get(boxe_class, default_color)
        xinf, yinf = min(x1, x2), min(y1, y2)
        w, h = abs(x2-x1), abs(y2-y1)
        rect = patches.Rectangle((xinf, yinf), w, h,
                                 linewidth=1, edgecolor=boxe_color,
                                 facecolor='none')
        if label_class:
            s = boxe_class
            confidence = bboxes.get("confidence", None)
            if confidence is not None:
                s += f": {confidence[i]*100:.1f}%"
            ax.text(xinf, yinf-1, s, color=boxe_color)
        ax.add_patch(rect)
    ax.set_xticks([])
    ax.set_yticks([])


def plot_matrix(table: pd.DataFrame,
                ax: Union[None, matplotlib.axes.Axes] = None,
                vmin: Optional[float] = None, vmax: Optional[float] = None,
                cmap: str = "viridis", color_bar: bool = False,
                format: str = ".3g", write_values: bool = False,
                **kwargs):
    """
    Plots the confusion matrix between prediction and target
    of a classifier

    Parameters
    ----------
    table : pd.DataFrame
        The matrix to plot the content of
    ax : None or matplotlib.axes.Axes
        The axis to draw on. If None, a new window is created.
    vmin : float or None
        min value for the colormap
    vmax : float or None
        max value for the colormap
    cmap : str
        The name of the maplotlib colormap
    color_bar : bool
        If True add a colorbar
    write_values : bool
        If True the value of each cell is writen
    format : str
        Formating used for the values writen in cells
    **kwargs : dict
        dict of additional keyword arguments passed to ax.text when writing
        values in cells
    """
    if ax is None:
        f, ax = plt.subplots()
    inf = vmin or table.min().min()
    sup = vmax or table.max().max()
    h = ax.imshow(table.to_numpy(), interpolation="nearest",
                  cmap=cmap, vmin=inf, vmax=sup)
    ax.grid(False)
    ax.set_xticks(range(len(table.columns)))
    ax.set_xticklabels(table.columns, rotation=45)
    ax.set_yticks(range(len(table.index)))
    ax.set_yticklabels(table.index, rotation=45)
    if color_bar:
        ax.get_figure().colorbar(h)
    if write_values:
        if isinstance(cmap, str):
            cmap = getattr(plt.cm, cmap)
        array = (h.get_array() - inf)/(sup - inf + 1.0E-8)
        brightness = np.mean(cmap(array)[:, :, :3], axis=-1)
        for y, cy in enumerate(table.index):
            for x, cx in enumerate(table.columns):
                val = table.loc[cy, cx]
                color = "white" if brightness[y, x] < 0.5 else "black"
                ax.text(x, y, f"{val:{format}}", va='center', ha='center',
                        color=color, **kwargs)



import torch
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint


class PaddedConv2d(torch.nn.Module):
    """
    PaddedConv2d is a wrapper around torch.nn.Conv2d that pads the input
    to maintain the same spatial dimension for the output after convolution.
    This layer is memory efficient thanks to gradient checkpointing.
    """

    def __init__(self, in_channels, out_channels, kernel_size,
                 stride=(1, 1), dilation=(1, 1), groups=1, bias=True,
                 device=None, dtype=None):
        """
        See torch.nn.Conv2d for parameters documentation
        """
        super().__init__()
        conv = torch.nn.Conv2d(in_channels, out_channels, kernel_size,
            stride=stride, dilation=dilation, groups=groups, bias=bias,
            device=device, dtype=dtype)
        kernel_size = tuple(k + (k-1)*(d-1) for k, d in zip(conv.kernel_size, conv.dilation))
        padding = [p for dim in kernel_size[::-1] for p in (dim//2, dim//2)]
        self.padding = torch.nn.ConstantPad2d(padding, 0.)
        self.conv = conv

    def _forward(self, X):
        return self.conv(self.padding(X))

    def forward(self, X: torch.Tensor):
        if self.training:
            X.requires_grad_(True)  # To ensure that the gradient is backpropagated in the convolution parameters
            return checkpoint(self._forward, X)
        else:
            return self._forward(X)



# class AdaptativePad2d(torch.nn.Module):
#     """
#     AdaptativePad2D is a layer that applies zero padding before a convolution
#     in order for the output feature map to have the same (height, width)
#     after the convolution as before padding
#     """

#     def __init__(self, kernel_size: tuple[int, int],
#                  stride: tuple[int, int], dilation: tuple[int, int]):
#         """
#         Parameters
#         ----------
#         kernel_size : tuple of (int, int)
#             the (height, width) of the convolution's convolved kernel
#         stride : tuple of (int, int)
#             the stride along (y, x) of the convolution
#         dilation : tuple of (int, int)
#             the dilation along (y, x) of the convolved kernel
#         """
#         super().__init__()
#         self.kernel = tuple(kernel_size)
#         self.stride = tuple(stride)
#         self.dilation = tuple(dilation)
    
#     def forward(self, X: torch.Tensor):
#         """
#         Parameters
#         ----------
#         X : torch.tensor
#             feature map of shape (N, C, H, W)
        
#         Returns
#         -------
#         torch.Tensor
#             tensor of shape (N, C, H+dh, W+dw) padded with zeros
#         """
#         _, _, h, w = X.shape
#         delta = [((L-1)*s - L + d*(k-1) + 1) for L, k, s, d
#                  in zip((h, w), self.kernel, self.stride, self.dilation)]
#         padding = [p for dim in delta[::-1] for p in (dim//2, dim//2)]
#         if any(p > 0 for p in padding):
#             return F.pad(X, padding, mode="constant", value=0.)
#         else:
#             return X